@@ -1,15 +1,16 @@
 package Weather::Underground;
 
 #
-# $Header: /cvsroot/weather::underground/Weather/Underground/Underground.pm,v 1.21 2003/10/31 05:14:44 mina Exp $
+# $Header: /cvsroot/weather::underground/Weather/Underground/Underground.pm,v 1.23 2003/11/13 03:29:26 mina Exp $
 #
 
 use strict;
-use vars qw($VERSION $CGI $CGIVAR $MYNAME $DEBUG);
+use vars qw($VERSION $CGI $CGIVAR $MYNAME $DEBUG %MODULES);
 use LWP::Simple qw($ua get);
 use HTML::TokeParser;
+use Fcntl qw(:flock);
 
-$VERSION = '2.11';
+$VERSION = '2.12';
 
 #
 # GLOBAL Variables Assignments
@@ -20,6 +21,17 @@ $CGIVAR = 'query';
 $MYNAME = "Weather::Underground";
 $DEBUG  = 0;
 
+%MODULES = (
+	"Data::Dumper" => 0,
+	"Storable"     => 0,
+	"FreezeThaw"   => 0,
+);
+
+foreach (keys %MODULES) {
+	eval { eval("require $_;") || die "$_ not found"; };
+	$MODULES{$_} = $@ ? 0 : 1;
+}
+
 =head1 NAME
 
 Weather::Underground - Perl extension for retrieving weather information from wunderground.com
@@ -29,20 +41,20 @@ Weather::Underground - Perl extension for retrieving weather information from wu
 	use Weather::Underground;
 
 	$weather = Weather::Underground->new(
-		place   =>      "Montreal, Canada",
-		debug           =>      0
+		place => "Montreal, Canada",
+		debug => 0,
 		)
 		|| die "Error, could not create new weather object: $@\n";
 
-	$arrayref = $weather->getweather()
-		|| die "Error, calling getweather() failed: $@\n";
+	$arrayref = $weather->get_weather()
+		|| die "Error, calling get_weather() failed: $@\n";
 
 	foreach (@$arrayref) {
 		print "MATCH:\n";
 		while (($key, $value) = each %{$_}) {
 			print "\t$key = $value\n";
-			}
 		}
+	}
 
 =head1 DESCRIPTION
 
@@ -50,92 +62,146 @@ Weather::Underground is a perl module which provides a simple OO interface to re
 
 =head1 CONSTRUCTOR
 
-	new(hashref);
+=over 4
 
-new() creates and returns a new Weather::Underground object.
+=item new(hash or hashref);
 
-"hashref" is a reference to a hash.
+Creates and returns a new Weather::Underground object.
+
+Takes either a hash (as the SYNOPSIS shows) or a hashref
 
 Required keys in the hash:
 
-	place
+=over 4
+
+=item place
+
+This key should be assigned the value of the geographical place you would like to retrieve the weather information for.  The format of specifying the place really depends on wunderground.com more than it depends on this perl module, however at the time of this writing they accept 'City', 'City, State', 'State', 'State, Country' and 'Country'.
+
+=back
 
 Optional keys in the hash:
 
-	debug
+=over 4
 
-"place" key should be assigned the value of the geographical place you would like to retrieve the weather information for.  The format of specifying the place really depends on wunderground.com more than it depends on this perl module, however at the time of this writing they accept 'City', 'City, State', 'State', 'State, Country' and 'Country'.
+=item cache_file
 
-"debug" key should be set to 0 or 1. 0 means no debugging information will be printed, 1 means debug information will be printed.
+This key should be assigned a file name to use as a cache.  The module will store and use data from that file instead of querying wunderground.com if cache_max_age has not been exceeded.
+
+This key is ignored if the cache_max_age key is not supplied.
+
+=item cache_max_age
+
+This key should be assigned a numeric value which is the number of seconds after which any data in the cache_file will be considered too old and a new request will be made to wunderground.com
+
+This key is ignored if the cache_file key is not supplied.
+
+=item debug
+
+This key should be set to a true or false false. A false value means no debugging information will be printed, a true value means debug information will be printed.
+
+=back
+
+=back
 
 =head1 METHODS
 
-	getweather();
+=over 4
 
-getweather() is used to initiate the connection to wunderground.com, query their system, and parse the results.
+=item get_weather()
 
-If no results are found, returns undef;
+This method is used to initiate the connection to wunderground.com, query their system, and parse the results or retrieve the results from the cache_file constructor key if appropriate.
+
+If no results are found, returns undef.
 
 If results are found, returns an array reference.  Each element in the array is a hash reference. Each hash contains information about a place that matched the query;
 
 Each hash contains the following keys:
 
-	place
-	(the exact place that was matched)
+=over 4
 
-	temperature_celsius
-	(the temperature in celsius)
+=item place
 
-	temperature_fahrenheit
-	(the temperature in fahrenheit)
+(the exact place that was matched)
 
-	humidity
-	(humidity percentage)
+=item temperature_celsius
 
-	conditions
-	(current sky, example: 'Partly cloudy')
+(the temperature in celsius)
 
-	wind
-	(wind direction and speed)
+=item temperature_fahrenheit
 
-	pressure
-	(the barometric pressure)
+(the temperature in fahrenheit)
 
-	windchill_celsius
-	(the temperature in celsius with wind chill considered)
+=item humidity
 
-	windchill_fahrenheit
-	(the temperature in fahrenheit with wind chill considered)
+(humidity percentage)
 
-	updated
-	(when the content was last updated on the server)
+=item conditions
+
+(current sky, example: 'Partly cloudy')
+
+=item wind
+
+(wind direction and speed)
+
+=item pressure
+
+(the barometric pressure)
+
+=item windchill_celsius
+
+(the temperature in celsius with wind chill considered)
+
+=item windchill_fahrenheit
+
+(the temperature in fahrenheit with wind chill considered)
+
+=item updated
+
+(when the content was last updated on the server)
+
+=back
+
+=back
 
 =head1 NOTICE
 
-1. Your query may result in more than 1 match. Each match is a hash reference added as a new value in the array which getweather() returns the reference to.
+=over 4
 
-2. Due to the differences between single and multiple-location matches, some of the keys listed above may not be available in multi-location matches.
+=item 1
+
+Your query may result in more than 1 match. Each match is a hash reference added as a new value in the array which get_weather() returns the reference to.
+
+=item 2
+
+Due to the differences between single and multiple-location matches, some of the keys listed above may not be available in multi-location matches.
+
+=back
 
 =head1 EXAMPLES
 
-Example 1: Print all matching information
+=over 4
+
+=item Example 1: Print all matching information
 
 	See SYNOPSIS
 
-Example 2: Print the Celsius temperature of the first matching place
+=item Example 2: Print the Celsius temperature of the first matching place
 
-        use Weather::Underground;
+	use Weather::Underground;
 
-        $weather = Weather::Underground->new(
-                place   =>      "Montreal",
-                debug           =>      0
-                )
-                || die "Error, could not create new weather object: $@\n";
+	$weather = Weather::Underground->new(
+		place   =>      "Montreal",
+		debug           =>      0
+		)
+		|| die "Error, could not create new weather object: $@\n";
 
-        $arrayref = $weather->getweather()
-                || die "Error, calling getweather() failed: $@\n";
+	$arrayref = $weather->get_weather()
+		|| die "Error, calling get_weather() failed: $@\n";
 
-	print "The celsius temperature at the first matching place is " . $arrayref->[0]->{temperature_celsius} . "\n";
+	print "The celsius temperature at $arrayref->[0]->{place} is $arrayref->[0]->{temperature_celsius}\n";
+
+=back
 
 =head1 ERRORS
 
@@ -156,12 +222,25 @@ Copyright (C) 2002-2003 Mina Naguib.  All rights reserved.  Use is subject to th
 =cut
 
 #
-# Public methods
+# Public methods:
 #
 
 sub new {
-	my ($class, %parameters) = @_;
+	my $class = shift;
 	my $self;
+	my %parameters;
+	my $module;
+	my $raw;
+	my $cache;
+	local (*FH);
+
+	if (ref($_[0]) eq "HASH") {
+		%parameters = %{ $_[0] };
+	}
+	else {
+		%parameters = @_;
+	}
+
 	$DEBUG = $parameters{debug};
 	_debug("Creating a new $MYNAME object");
 	if (!$parameters{place}) {
@@ -169,14 +248,120 @@ sub new {
 		return undef;
 	}
 	$self = {
-		_place => $parameters{place},
-		_url   => $CGI . '?' . $CGIVAR . '=' . $parameters{place}
+		"place" => $parameters{place},
+		"_url"  => $CGI . '?' . $CGIVAR . '=' . $parameters{place}
 	};
+	if ($parameters{cache_max_age} && $parameters{cache_file}) {
+
+		#
+		# We've been requested to use caching - let's do sanity, then populate $module and $cache
+		#
+		if (!grep { $_ } values %MODULES) {
+			_debug("Error: Can not use cache_file when none of the needed serialization modules (" . join(" or ", keys %MODULES) . ") are installed");
+			return undef;
+		}
+		if ($parameters{cache_max_age} !~ /^[0-9.]+$/) {
+			_debug("Error: Supplied cache_max_age key must be a number");
+			return undef;
+		}
+		if (-f $parameters{cache_file}) {
+
+			#
+			# The cache file already exists
+			#
+			if (!open(FH, $parameters{cache_file})) {
+				_debug("Error: Failed to open $parameters{cache_file} for reading: $!");
+				return undef;
+			}
+			if (!flock(FH, LOCK_EX)) {
+				close(FH);
+				_debug("Error: Failed to obtain an exclusive lock on $parameters{cache_file}: $!");
+				return undef;
+			}
+			if (!seek(FH, 0, 0)) {
+				flock(FH, LOCK_UN);
+				close(FH);
+				_debug("Error: Failed to seek to the beginning of $parameters{cache_file}: $!");
+				return undef;
+			}
+			$module = <FH>;
+			chomp $module;
+			if (!exists $MODULES{$module}) {
+				flock(FH, LOCK_UN);
+				close(FH);
+				_debug("cache_file $parameters{cache_file} does not appear to be a valid Weather::Underground cache file");
+				return undef;
+			}
+			elsif (!$MODULES{$module}) {
+				flock(FH, LOCK_UN);
+				close(FH);
+				_debug("cache_file $parameters{cache_file} with serialization module $module which is not installed on this machine.  Please install it or delete the cache file to start with a fresh one");
+				return undef;
+			}
+
+			$cache = "";
+			$raw   = "";
+			while (<FH>) {
+				$raw .= $_;
+			}
+			flock(FH, LOCK_UN);
+			close(FH);
+
+			#
+			# Now deserialize $cache
+			#
+			if ($module eq "Data::Dumper") {
+				my $VAR1;
+				$cache = eval($raw);
+			}
+			elsif ($module eq "Storable") {
+				$cache = Storable::thaw($raw);
+			}
+			elsif ($module eq "FreezeThaw") {
+				$cache = FreezeThaw::thaw($raw);
+			}
+
+			if (ref($cache) ne "HASH") {
+				_debug("Failed to deserialize cache with module $module - [$!] [$@] got non-hashref [$cache] from raw [$raw]");
+				return undef;
+			}
+		}
+		else {
+
+			#
+			# The cache file does not exist - create new one
+			#
+			if (!open(FH, ">$parameters{cache_file}")) {
+				_debug("Error: Failed to open $parameters{cache_file} for writing: $!");
+				return undef;
+			}
+			close(FH);
+			$module = (sort grep { $MODULES{$_} } keys %MODULES)[0];
+			$cache = {};
+		}
+
+		#
+		# If we've reached here, cache_file and cache_max_age are good
+		#
+		$self->{cache_file}    = $parameters{cache_file};
+		$self->{cache_max_age} = $parameters{cache_max_age};
+		$self->{_cache_module} = $module;
+		$self->{_cache_cache}  = $cache;
+	}
+	elsif ($parameters{cache_max_age} || $parameters{cache_file}) {
+		_debug("cache_max_age or cache_file was supplied without the other - ignoring it");
+	}
+
 	bless($self, $class);
 	return $self;
 }
 
+# legacy:
 sub getweather {
+	return get_weather(@_);
+}
+
+sub get_weather {
 	my ($self) = @_;
 	my $document;
 	my $parser;
@@ -185,8 +370,27 @@ sub getweather {
 	my $text;
 	my $arrayref = [];
 	my $oldagent;
+	local (*FH);
+
+	_debug("Getting weather info for " . $self->{place});
+
+	if ($self->{_cache_cache}) {
+
+		#
+		# We have a cache
+		#
+		_debug("Checking cache");
+		if (exists $self->{_cache_cache}->{ $self->{place} }) {
+			if ((time - $self->{_cache_cache}->{ $self->{place} }->{"time"}) <= $self->{cache_max_age}) {
+				_debug("Found in cache within cache_max_age");
+				return $self->{_cache_cache}->{ $self->{place} }->{"arrayref"};
+			}
+			else {
+				_debug("Found in cache but too old");
+			}
+		}
+	}
 
-	_debug("Getting weather info for " . $self->{_place});
 	_debug("Retrieving url " . $self->{_url});
 
 	$oldagent = $ua->agent();
@@ -240,6 +444,7 @@ sub getweather {
 					#
 					# This is the title for a header column - store it for later use when encountering content under same column
 					#
+					_debug("Header text read [$text]");
 					$state{"header_$state{headernumber}"} = uc($text);
 				}
 				elsif ($state{"incontent"}) {
@@ -247,6 +452,7 @@ sub getweather {
 					#
 					# This is content we're interested in - store it under the header title of the same column number
 					#
+					_debug("Content text read [$text]");
 					$state{ "content_" . $state{ "header_" . $state{"contentnumber"} } } = $text;
 				}
 			}
@@ -255,6 +461,8 @@ sub getweather {
 				#
 				# A new cell in the header of the table that has the info we need has started
 				#
+				_debug("A new table header cell started");
+				_debug("This means we entered the interesting table") unless $state{"intable"};
 				$state{"headernumber"}++;
 				$state{"inheader"}  = 1;
 				$state{"intable"}   = 1;
@@ -270,6 +478,7 @@ sub getweather {
 					#
 					# This is the end of the header and the beginning of the content rows
 					#
+					_debug("A new row started while we're in header - assuming end of header and upcoming content");
 					$state{"inheader"}      = 0;
 					$state{"incontent"}     = 1;
 					$state{"contentnumber"} = 0;
@@ -279,6 +488,7 @@ sub getweather {
 					#
 					# This is a new content row beginning
 					#
+					_debug("A new row containing content started");
 					$state{"contentnumber"} = 0;
 
 					#
@@ -288,6 +498,13 @@ sub getweather {
 						delete $state{$_} if /^content_/;
 					}
 				}
+				else {
+
+					#
+					# Shouldn't reach here
+					#
+					_debug("Shouldn't see this - new row started while we're in table but we're not in header or content!");
+				}
 			}
 			elsif ($token->[0] eq "E" && uc($token->[1]) eq "TR" && $state{"incontent"}) {
 
@@ -296,6 +513,7 @@ sub getweather {
 				#
 				# Save the data
 				#
+				_debug("End of content row");
 				_state2result(\%state, $arrayref);
 
 			}
@@ -305,6 +523,7 @@ sub getweather {
 				#
 				# The beginning of a new cell with content
 				#
+				_debug("Beginning of new content cell");
 				$state{"contentnumber"}++;
 			}
 			elsif ($token->[0] eq "E" && uc($token->[1]) eq "TABLE" && $state{"intable"}) {
@@ -312,6 +531,7 @@ sub getweather {
 				#
 				# The table that has the data is finished - no need to keep parsing
 				#
+				_debug("End of table while we're in table - we're done");
 				last;
 			}
 		}
@@ -323,6 +543,8 @@ sub getweather {
 		#
 		# We use single-location algorithm
 		#
+		_debug("Single-location result detected");
+
 		while ($token = $parser->get_token) {
 			if ($token->[0] eq "T" && !$token->[2]) {
 
@@ -337,11 +559,12 @@ sub getweather {
 				$text =~ s/\s+$//g;
 				$text =~ s/\s+/ /g;
 
-				if (uc($text) eq "CONDITIONS" && !$state{"intable"}) {
+				if ($text =~ /^(current\s+)?conditions$/i && !$state{"intable"}) {
 
 					#
 					# We just entered the table that has the data we want
 					#
+					_debug("Matched conditions text - entered interesting table");
 					$state{"intable"}     = 1;
 					$state{"intopheader"} = 0;
 					$state{"incontent"}   = 0;
@@ -351,6 +574,7 @@ sub getweather {
 					#
 					# This is the top header
 					#
+					_debug("Matched top header text");
 					($state{"content_UPDATED"}) = ($text =~ /updated\s*:?\s*(.+?)\s*observ|\Z/i);
 					($state{"content_PLACE"})   = ($text =~ /observed\s+at\s+:?\s*(.+)/i);
 				}
@@ -364,6 +588,7 @@ sub getweather {
 						#
 						# It's a header - remember to associate the upcoming content under it
 						#
+						_debug("Read header text [$text]");
 						$state{"header"} = uc($text);
 					}
 					else {
@@ -371,6 +596,7 @@ sub getweather {
 						#
 						# It's a content - associate it with the previous header
 						#
+						_debug("Read content text [$text]");
 						$state{ "content_" . $state{"header"} } = $text;
 					}
 				}
@@ -385,6 +611,7 @@ sub getweather {
 					#
 					# Should never reach here - but in case we do, we're about to start the top header
 					#
+					_debug("Should not have reached here (new row while in table but not in top header or in content!)");
 					$state{"intopheader"} = 1;
 				}
 				elsif ($state{"intopheader"}) {
@@ -392,6 +619,7 @@ sub getweather {
 					#
 					# The top header is finished and the content is coming up
 					#
+					_debug("New row while in top header - assuming top header ending and content coming up");
 					$state{"intopheader"} = 0;
 					$state{"incontent"}   = 1;
 				}
@@ -400,6 +628,7 @@ sub getweather {
 					#
 					# A new header+content coming up
 					#
+					_debug("New content row starting");
 					$state{"contentnumber"} = 0;
 				}
 			}
@@ -408,6 +637,7 @@ sub getweather {
 				#
 				# A new header or content cell is starting
 				#
+				_debug("New header or content cell starting");
 				$state{"contentnumber"}++;
 			}
 			elsif ($token->[0] eq "E" && uc($token->[1]) eq "TABLE" && $state{"intable"}) {
@@ -415,6 +645,7 @@ sub getweather {
 				#
 				# Done parsing - save the data
 				#
+				_debug("Table closed while in table - end of interesting data");
 				_state2result(\%state, $arrayref);
 
 				#
@@ -430,6 +661,47 @@ sub getweather {
 		return undef;
 	}
 	else {
+		if ($self->{cache_file}) {
+
+			#
+			# Let's save the result into the cache_file before we return it
+			#
+			_debug("Saving results into cache_file $self->{cache_file}");
+
+			$self->{_cache_cache}->{ $self->{place} } = {
+				"time"     => time,
+				"arrayref" => $arrayref,
+			};
+			if (open(FH, ">$self->{cache_file}")) {
+				if (flock(FH, LOCK_EX)) {
+					if (seek(FH, 0, 0)) {
+						print FH $self->{_cache_module};
+						print FH "\n";
+						if ($self->{_cache_module} eq "Data::Dumper") {
+							print FH Data::Dumper::Dumper($self->{_cache_cache});
+						}
+						elsif ($self->{_cache_module} eq "Storable") {
+							print FH Storable::freeze($self->{_cache_cache});
+						}
+						elsif ($self->{_cache_module} eq "FreezeThaw") {
+							print FH FreezeThaw::freeze($self->{_cache_cache});
+						}
+						flock(FH, LOCK_UN);
+						close(FH);
+					}
+					else {
+						_debug("Error: Failed to seek to beginning of cache_file $self->{cache_file}: $!");
+					}
+				}
+				else {
+					_debug("Error: Failed to lock cache_file $self->{cache_file} exclusively: $!");
+				}
+			}
+			else {
+				_debug("Error: Failed to open cache_file $self->{cache_file} for writing: $!");
+			}
+
+		}
 		return $arrayref;
 	}
 
